pub mod binary;
pub mod broadcast;
pub mod cast;
pub mod change_axes;
pub mod concat;
pub mod element_wise;
pub mod gemm;
pub mod konst;
pub mod new_gelu;
pub mod reduce;
pub mod rms_norm;
pub mod silu;
pub mod slice;
pub mod softmax;
pub mod sync;

pub use binary::MetalBinOp;
pub use broadcast::MetalMultiBroadcastTo;
pub use cast::MetalCast;
pub use change_axes::{MetalAxisOp, MetalIntoShape};
pub use concat::MetalConcat;
pub use element_wise::MetalElementWiseOp;
pub use gemm::MetalGemm;
pub use konst::MetalConst;
pub use new_gelu::MetalNewGelu;
pub use reduce::MetalReduce;
pub use rms_norm::MetalRmsNorm;
pub use silu::MetalSilu;
pub use slice::MetalSlice;
pub use softmax::MetalSoftmax;
pub use sync::{MetalSync, MetalSyncKind};
