#![allow(non_upper_case_globals)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(improper_ctypes)]
#![allow(deref_nullptr)]
#![allow(clippy::redundant_static_lifetimes)]
#![allow(clippy::useless_transmute)]

include!("../bindings.rs");
